/*
 * memory.c -- memory manager
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "log.h"
#include "memory.h"


#undef DEBUG


/*
 * Memory is managed in pieces called "chunks". A chunk consists of
 * some management information followed by a plain byte array which
 * is handed over to the user.
 * The size of the byte array is always a power of 2, chosen large
 * enough so that the size requested by the user is less than or
 * equal to that power.
 * Every chunk is either linked within the chunk table or the cluster
 * table. The chunk table is ordered by size and has one entry per
 * power of two. The chain of chunks linked by entry n holds all
 * free chunks of size 2^n. The cluster table is ordered by cluster
 * number and has one entry per cluster number. The chain of chunks
 * linked by entry n holds all allocated chunks of cluster n. These
 * chunks are possibly of different sizes and are preferrably freed
 * together with a single call to memoryFreeCluster(), although it
 * is possible to free a single chunk alone.
 * Allocation takes place from the appropriate chain of free chunks.
 * If this chain is empty, a new chunk of this size is requested
 * from the C runtime system by calling malloc().
 */


typedef struct chunk {
  struct chunk *next;	/* link to next free chunk or next chunk in cluster */
  int logSize;		/* log2(size) of this chunk; this is never changed */
  int cluster;		/* cluster this chunk belongs to; only if allocated */
  /* user accessible area follows */
} Chunk;


static int chunkTableSize;
static Chunk **chunkTable;

static int clusterTableSize;
static Chunk **clusterTable;


/*
 * This function must be called to initialize the memory manager.
 * It needs the log2 of the maximum size of any chunk ever expected
 * to be allocated (rounded up to the next integer) as well as the
 * number of "clusters", i.e. regions of memory which are expected
 * to be freed at the same time.
 */
void memoryInit(int logMaxSize, int numClusters) {
  int i;

#ifdef DEBUG
  logMessage("memoryInit: max size = %u, max cluster number = %d\n",
             (Uint) 1 << logMaxSize, numClusters - 1);
#endif
  chunkTableSize = logMaxSize + 1;
  chunkTable = malloc(chunkTableSize * sizeof(Chunk *));
  if (chunkTable == NULL) {
    logPanic("cannot allocate memory chunk table");
  }
  for (i = 0; i < chunkTableSize; i++) {
    chunkTable[i] = NULL;
  }
  clusterTableSize = numClusters;
  clusterTable = malloc(clusterTableSize * sizeof(Chunk *));
  if (clusterTable == NULL) {
    logPanic("cannot allocate memory cluster table");
  }
  for (i = 0; i < clusterTableSize; i++) {
    clusterTable[i] = NULL;
  }
}


/*
 * This function returns all memory to the C runtime system.
 */
void memoryExit(void) {
  int i;
  Chunk *chunk;

  for (i = 0; i < clusterTableSize; i++) {
    while (clusterTable[i] != NULL) {
      chunk = clusterTable[i];
      clusterTable[i] = chunk->next;
      free(chunk);
#ifdef DEBUG
      logMessage("memoryExit: chunk freed at address = %p\n",
                 chunk);
#endif
    }
  }
  free(clusterTable);
  for (i = 0; i < chunkTableSize; i++) {
    while (chunkTable[i] != NULL) {
      chunk = chunkTable[i];
      chunkTable[i] = chunk->next;
      free(chunk);
#ifdef DEBUG
      logMessage("memoryExit: chunk freed at address = %p\n",
                 chunk);
#endif
    }
  }
  free(chunkTable);
}


/*
 * This function is called whenever the user wants to allocate
 * memory. He must supply the requested size and a cluster number,
 * which serves to group allocated memory according to lifetime.
 */
void *memoryAllocate(Uint size, int cluster) {
  int logSize;
  Chunk *chunk;

#ifdef DEBUG
  logMessage("memoryAllocate: requested size = %u, cluster = %d\n",
             size, cluster);
#endif
  if (size == 0) {
    logPanic("memory allocation request with size 0 not allowed");
  }
  for (logSize = 0; logSize < chunkTableSize; logSize++) {
    if (((Uint) 1 << logSize) >= size) {
      break;
    }
  }
  if (logSize == chunkTableSize) {
    logPanic("memory allocation request exceeds size limit");
  }
  if (cluster < 0 || cluster >= clusterTableSize) {
    logPanic("memory allocation request with illegal cluster number");
  }
#ifdef DEBUG
  logMessage("memoryAllocate: log2(size) = %d, size = %u\n",
             logSize, (Uint) 1 << logSize);
#endif
  if (chunkTable[logSize] == NULL) {
#ifdef DEBUG
    logMessage("memoryAllocate: malloc(2^%d)\n",
               logSize);
#endif
    chunk = malloc(sizeof(Chunk) + ((Uint) 1 << logSize));
    if (chunk == NULL) {
      logPanic("out of memory");
    }
#ifdef DEBUG
    logMessage("memoryAllocate: chunk address = %p\n",
               chunk);
#endif
    chunk->logSize = logSize;
    chunk->next = chunkTable[logSize];
    chunkTable[logSize] = chunk;
  }
  chunk = chunkTable[logSize];
  chunkTable[logSize] = chunk->next;
  chunk->cluster = cluster;
  chunk->next = clusterTable[cluster];
  clusterTable[cluster] = chunk;
#ifdef DEBUG
  logMessage("memoryAllocate: void address = %p\n",
             (Uchar *) chunk + sizeof(Chunk));
#endif
  return (void *) ((Uchar *) chunk + sizeof(Chunk));
}


/*
 * This function frees a particular chunk of memory. This can be
 * done, although it is more efficient to free a whole cluster
 * with the function memoryFreeCluster(); see below.
 */
void memoryFree(void *p) {
  Chunk *chunk;
  int logSize;
  int cluster;
  Chunk *prev;

#ifdef DEBUG
  logMessage("memoryFree: void address = %p\n",
             p);
#endif
  if (p == NULL) {
    logPanic("memory free with NULL argument not allowed");
  }
  chunk = (Chunk *) ((Uchar *) p - sizeof(Chunk));
#ifdef DEBUG
  logMessage("memoryFree: chunk address = %p\n",
             chunk);
#endif
  logSize = chunk->logSize;
  cluster = chunk->cluster;
#ifdef DEBUG
  logMessage("memoryFree: log2(size) = %d, cluster = %d\n",
             logSize, cluster);
#endif
  if (clusterTable[cluster] == chunk) {
    clusterTable[cluster] = chunk->next;
  } else {
    prev = clusterTable[cluster];
    while (prev->next != chunk) {
      prev = prev->next;
    }
    prev->next = chunk->next;
  }
  chunk->next = chunkTable[logSize];
  chunkTable[logSize] = chunk;
}


/*
 * This function frees a whole cluster of memory chunks.
 */
void memoryFreeCluster(int cluster) {
  Chunk *chunk;
  int logSize;

  if (cluster < 0 || cluster >= clusterTableSize) {
    logPanic("memory free cluster with illegal cluster number");
  }
  while (clusterTable[cluster] != NULL) {
    chunk = clusterTable[cluster];
    clusterTable[cluster] = chunk->next;
    logSize = chunk->logSize;
    chunk->next = chunkTable[logSize];
    chunkTable[logSize] = chunk;
#ifdef DEBUG
    logMessage("memoryFreeCluster: cluster = %d, "
               "chunk freed at address = %p\n",
               cluster, chunk);
#endif
  }
}
